# Imports
import numpy as np
np.random.seed(0)
import pandas as pd
import subprocess

from sklearn.preprocessing import StandardScaler
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix, classification_report

import torch
torch.manual_seed(0)
from torch import nn, optim
import torch.nn.functional as F
if torch.cuda.is_available():
    print(torch.cuda.get_device_name(torch.cuda.current_device()))
    torch.cuda.empty_cache()

import warnings
warnings.filterwarnings('ignore')

# Neural network architecture
class Net(nn.Module):
    def __init__(self, n_features):
        super(Net, self).__init__()

        self.fc1 = nn.Linear(n_features, 16)
        self.fc2 = nn.Linear(16, 4)
        self.fc3 = nn.Linear(4, 1)

    def forward(self, x):
        x = F.relu(self.fc1(x))
        x = F.relu(self.fc2(x))
        return torch.sigmoid(self.fc3(x))

def main():
    print("\nData processing...")

    # Import dataset
    df = pd.read_csv('weatherAUS.csv')
    print("Dataset shape:", df.shape)
    print("Proportions:")
    print(df.RainTomorrow.value_counts() / df.shape[0])

    # Drop RISK_MM variable as indicated
    df.drop(['RISK_MM'], axis=1, inplace=True)

    # Process Date
    df['Date'] = pd.to_datetime(df['Date'])
    df['Year'] = df['Date'].dt.year
    df['Month'] = df['Date'].dt.month
    df['Day'] = df['Date'].dt.day
    df.drop(['Date'], axis=1, inplace=True)

    # Process rain
    df['RainToday'].replace({'No': 0, 'Yes': 1}, inplace = True)
    df['RainTomorrow'].replace({'No': 0, 'Yes': 1}, inplace = True)


    # Create train and test dataset
    print("\nCreating train and test dataset...")
    X = df.drop(['RainTomorrow'], axis=1).copy()
    y = df['RainTomorrow'].copy()

    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size = 0.25, random_state = 0)
    print("Train set's shape:", X_train.shape)
    print("Test set's shape:", X_test.shape)

    # Replace missing values with the column median for numerical values,
    # and the most given category for the categorical ones
    numerical_columns = [col for col in X.columns if X[col].dtypes != 'O']
    categorical_columns = [col for col in X.columns if X[col].dtypes == 'O']

    for df1 in [X_train, X_test]:
        for col in numerical_columns:
            col_median = X_train[col].median()
            df1[col].fillna(col_median, inplace=True)
        df1['WindGustDir'].fillna(X_train['WindGustDir'].mode()[0], inplace=True)
        df1['WindDir9am'].fillna(X_train['WindDir9am'].mode()[0], inplace=True)
        df1['WindDir3pm'].fillna(X_train['WindDir3pm'].mode()[0], inplace=True)

    # Process categorical values
    X_train = pd.get_dummies(X_train, columns=categorical_columns)
    X_test = pd.get_dummies(X_test, columns=categorical_columns)

    # Standardization
    scaler = StandardScaler()
    X_train = scaler.fit_transform(X_train)
    X_test = scaler.transform(X_test)


    ### NEURAL NETWORK ###
    # Pre-processing
    print("\nSetting pytorch variables...")


    # Transform arrays to Tensors
    X_train = torch.from_numpy(X_train).float()
    y_train = torch.squeeze(torch.from_numpy(y_train.to_numpy()).float())

    X_test = torch.from_numpy(X_test).float()
    y_test = torch.squeeze(torch.from_numpy(y_test.to_numpy()).float())

    print(X_train.shape, y_train.shape)
    print(X_test.shape, y_test.shape)

    # Model
    net = Net(X_train.shape[1])
    print("Neural network:", net)

    # Binary Cross Entropy loss
    loss_function = nn.BCELoss()

    # Adam optimizer
    optimizer = optim.Adam(net.parameters(), lr=0.001)

    # If GPU available, transfer calculation to GPU
    device = torch.device("cuda:1" if torch.cuda.is_available() else "cpu")
    print("Running on", device)

    if torch.cuda.is_available():
        # If parallel execution - selected GPUs in device_ids
        # net = nn.DataParallel(net, device_ids=[1, 2, 3])

        # Transfers to GPUs
        X_train = X_train.to(device)
        y_train = y_train.to(device)
        X_test = X_test.to(device)
        y_test = y_test.to(device)
        net = net.to(device)
        loss_function = loss_function.to(device)


    # TRAINING
    print("\nTraining...")

    # CPU and GPU usage in a log file
    with open('/job/cpu.txt', 'w') as fcpu:
        proc = subprocess.Popen(["top", "-n", "60", "-b"], stdout=fcpu)

    if torch.cuda.is_available():
        with open('/job/gpu.txt', 'w') as fgpu:
            proc2 = subprocess.Popen(["nvidia-smi", "--loop-ms=200"], stdout=fgpu)

    for epoch in range(1001):
        y_pred = net(X_train)
        y_pred = torch.squeeze(y_pred)

        # Compute loss
        train_loss = loss_function(y_pred, y_train)

        # Optimization
        optimizer.zero_grad()
        train_loss.backward()
        optimizer.step()

        # Print loss and accuracies
        if epoch % 100 == 0:
            if torch.cuda.is_available():
                print('Memory Usage:')
                print('Allocated:', round(torch.cuda.memory_allocated(device)/1024**3,3), 'GB')
                print('Cached:   ', round(torch.cuda.memory_reserved(device)/1024**3,3), 'GB')

            with torch.no_grad():
                train_acc = (y_train == y_pred.ge(.5).view(-1)).sum().float() / len(y_train)
                y_test_pred = net(X_test)
                y_test_pred = torch.squeeze(y_test_pred)
                test_loss = loss_function(y_test_pred, y_test)
                test_acc = (y_test == y_test_pred.ge(.5).view(-1)).sum().float() / len(y_test)
                print(
                    f'''Epoch {epoch}
                    Train set - loss: {train_loss:.5f}, accuracy: {train_acc:.4f}
                    Test  set - loss: {test_loss:.5f}, accuracy: {test_acc:.4f}
                    ''')

    # stop recording CPU and GPU usage
    proc.kill()
    if torch.cuda.is_available():
        proc2.kill()
    
    ### SAVE RESULTS ###
    print("\nSaving results...")
    # Save model
    torch.save(net, 'rain_model.pth')

    # Prediction
    with torch.no_grad():
        y_pred_f = net(X_test)
        y_pred_f = y_pred_f.ge(.5).view(-1)

    # Save predictions
    np.savetxt('y_pred', y_pred_f.detach().cpu().numpy())

    # Evaluation
    print("Classification report:")
    print(classification_report(y_test.detach().cpu().numpy(), y_pred_f.detach().cpu().numpy(),
                                target_names=['No Rain', 'Rain']))

    cm_matrix = pd.DataFrame(data=confusion_matrix(y_test.detach().cpu().numpy(), y_pred_f.detach().cpu().numpy()),
                             columns=['No Rain', 'Rain'],
                             index=['No Rain', 'Rain'])
    print("Confusion matrix:")
    print(cm_matrix/cm_matrix.sum(axis=1)[:, np.newaxis])

if __name__ == '__main__':
    main()
